"""
Decorator utilities
"""
import functools

from requests.exceptions import ConnectionError as RequestsConnectionError


try:
    from functools import cached_property
except ImportError:  # pragma: no cover
    # Python 3.7
    from cached_property import cached_property


def retry_on_connection_error(func=None, max_retries=5):
    """
    Decorator to automatically retry a function when a `ConnectionError` (such
    as a Broken Pipe error) is raised.
    """
    def decorate_func(f):
        @functools.wraps(f)
        def new_func(*args, **kwargs):
            retries = 0
            while retries < max_retries:
                try:
                    return f(*args, **kwargs)
                except RequestsConnectionError:
                    retries += 1
                except Exception:
                    # Unexpected error, raise the message so it shows up
                    raise

            raise Exception('Maximum retries exceeded')

        return new_func

    return decorate_func(func) if func else decorate_func



"""
Common test fixtures and utilities
"""
import sys
from typing import Union, List

import pytest

from wystia.utils.parse import as_list


# Check if system Python version is 3.6 or greater
PY36 = sys.version_info[0:2] >= (3, 6)


@pytest.fixture(scope='session')
def mock_video_id():
    """Returns a dummy hashed video ID for testing purposes."""
    return 'abc-01234567'


@pytest.fixture(scope='session')
def mock_project_id():
    """Returns a dummy hashed project ID for testing purposes."""
    return 'xyz-76543210'


@pytest.fixture(scope='session')
def mock_api_token():
    """Returns a dummy Wistia API token for testing purposes."""
    return 'abc-xyz-123456789'


@pytest.fixture(scope='session')
def mock_file_name():
    """Returns a dummy file path for testing purposes."""
    return 'abc-1234567.mp4'


class TestsWithMarkSkipper:
    """
    Util to skip tests with mark, unless cli option provided.
    """
    def __init__(self, test_marks: Union[str, List[str]],
                 cli_option_name: str,
                 cli_option_help: str):

        self.test_marks = as_list(test_marks)
        self.cli_option_name = cli_option_name
        self.cli_option_help = cli_option_help

    def pytest_addoption_hook(self, parser):
        parser.addoption(
            self.cli_option_name,
            action='store_true',
            default=False,
            help=self.cli_option_help,
        )

    def pytest_runtest_setup(self, item):
        if any(mark in item.keywords for mark in self.test_marks) \
                and not item.config.getoption(self.cli_option_name):
            self._skip_test()

    def _skip_test(self):
        reason = 'need {} option to run this test'.format(self.cli_option_name)
        pytest.skip(reason)


mark_skipper = TestsWithMarkSkipper(
    test_marks=['mutative', 'long'],
    cli_option_name="--run-all",
    cli_option_help="run all test cases, including any potentially "
                    "destructive tests",
)


pytest_addoption = mark_skipper.pytest_addoption_hook
pytest_runtest_setup = mark_skipper.pytest_runtest_setup



from __future__ import annotations

from requests import Session

from .api_base import _BaseWistiaApi
from .config import WistiaConfig
from .errors import NoSuchMedia
from .models import VideoEmbedData


class WistiaEmbedApi(_BaseWistiaApi):
    """
    Helper class to interact with the Wistia Media Embed API (docs below)
      https://wistia.com/support/developers/embed-options

    """
    _API_ENDPOINT = WistiaConfig.EMBED_URL

    @classmethod
    def _get_session(cls, additional_status_force_list=None) -> Session:
        """
        Override to not wrap with the decorator that increments the
        :attr:`_REQUEST_COUNT` attribute, since it's not confirmed whether the
        Wistia Data API rate limitations apply in the case of this endpoint.
        """
        return cls._create_session(additional_status_force_list)

    @classmethod
    def get_data(cls, video_id: str) -> VideoEmbedData:
        """
        Get media embed data for a Wistia video using the endpoint to the
        `.jsonp` file

        """
        r = cls.session().get(
            WistiaConfig.MEDIAS_EMBED_URL.format(media_id=video_id))
        r.raise_for_status()

        data = r.json()
        if 'error' in data:
            # Wistia Embed response contains an error object like below:
            #   {'error': True, 'iframe': True}
            raise NoSuchMedia(video_id)

        return VideoEmbedData.from_dict(data.get('media', {}))

    @classmethod
    def asset_url(cls, video_id: str | None = None,
                  media_data: VideoEmbedData | None = None,
                  asset_type='original') -> str | None:
        """
        Get the media asset url stored on Wistia, by default for the
        "original" video.

        Note: one of `video_id` or `media_data` must be specified.

        Note that Wistia also has separate asset url's for various resolutions
        on each video.

        """
        if media_data is None:
            media_data = cls.get_data(video_id)

        mp4_url = None
        for asset in media_data.assets:
            if asset.type == asset_type:
                url = asset.url
                mp4_url = url.replace('.bin', '/file.mp4', 1)

        return mp4_url

    @classmethod
    def num_assets(cls, video_id: str | None = None,
                   media_data: VideoEmbedData | None = None,
                   asset_type='mp4_alternate_audio') -> int:
        """
        Return the total number of assets (by default, AD files) associated
        with a Wistia video.

        """
        if media_data is None:
            media_data = cls.get_data(video_id)

        asset_count = 0
        for asset in media_data.assets:
            if asset.type == asset_type:
                asset_count += 1

        return asset_count



"""
Utility functions for interacting with the Wistia API.
"""
from __future__ import annotations

from requests import HTTPError

from .api_data import WistiaDataApi
from .config import WistiaConfig
from .errors import NoSuchProject
from .log import LOG
from .models import (Container, Project, Video,
                     Customizations, Plugin, CaptionsV1)


class WistiaHelper:
    """
    Helper class for interacting with Wistia API and Wistia videos, which
    further simplifies any calls. This assumes that :class:`WistiaApi` has
    been configured as needed with the API token.

    """

    @staticmethod
    def is_archived_video(video_name: str) -> bool:
        """
        Check if this is an archived video which is automatically created by
        Wistia.

        In that case, we shouldn't order captions for it, since it's a copy of
        a video that was replaced with new media.

        Source: https://wistia.com/learn/product-updates/improved-library-management-tools  # noqa: E501
        """
        return '[Archived' in video_name and video_name[-1] == ']'

    @classmethod
    def video_exists(cls, video_id: str) -> bool:
        """
        Check if a video exists on Wistia.
        """
        status = WistiaDataApi.session().head(
            WistiaConfig.MEDIAS_SHOW_URL.format(media_id=video_id)
        ).status_code

        return not (status == 404)

    @classmethod
    def update_video_name(
        cls,
        video_id: str,
        video_title: str
    ) -> Video:
        """Update the title for a Wistia video."""
        return WistiaDataApi.update_video(video_id, video_title)

    @classmethod
    def has_captions_enabled(cls, video_id: str) -> bool:
        """
        Check if a video has captions available in the player settings.

        :return: A boolean indicating if the video player has captions
          available.
        """
        customizations = WistiaDataApi.get_customizations(video_id)

        return customizations.plugin.captions_v1.on is True

    @classmethod
    def enable_captions(
        cls,
        video_id: str,
        on_by_default: bool = False
    ):
        """
        Enable captions on a Wistia video.
        """
        data = Customizations(
            plugin=Plugin(
                captions_v1=CaptionsV1(
                    on_by_default=on_by_default,
                    on=True
                )
            )
        )

        return WistiaDataApi.update_customizations(video_id, data)

    @classmethod
    def enable_ad(cls, video_id: str):
        """
        Enable audio descriptions on a Wistia video.
        """
        data = Customizations(audio_description_is_required=True)

        return WistiaDataApi.update_customizations(video_id, data)

    @classmethod
    def enable_captions_and_ad(
        cls,
        video_id: str,
        on_by_default: bool = False
    ):
        """
        Enable captions and AD on a Wistia video.
        """
        data = Customizations(
            plugin=Plugin(
                captions_v1=CaptionsV1(
                    on_by_default=on_by_default,
                    on=True
                )
            ),
            audio_description_is_required=True
        )

        return WistiaDataApi.update_customizations(video_id, data)

    @classmethod
    def disable_captions_and_ad(
        cls,
        video_id: str,
        on_by_default: bool = False
    ):
        """
        Disable captions and AD on a Wistia video.
        """
        data = Customizations(
            plugin=Plugin(
                captions_v1=CaptionsV1(
                    on_by_default=on_by_default,
                    on=False
                )
            ),
            audio_description_is_required=False
        )

        return WistiaDataApi.update_customizations(video_id, data)

    @classmethod
    def customize_video_on_wistia(
        cls,
        video_id: str,
        player_color: str
    ):
        """Set commonly used customization options for a media on Wistia."""
        customizations = {'playerColor': player_color}

        r = WistiaDataApi.session().put(
            WistiaConfig.CUSTOMIZATION_URL.format(media_id=video_id),
            json=customizations
        )

        try:
            r.raise_for_status()
        except HTTPError:
            LOG.error('Wistia Customizations API. status=%d text=%s',
                      r.status_code, r.text)
            raise

        LOG.info('[%s] Wistia: Customize Video success', r.elapsed)
        return r.json()

    @classmethod
    def project_details(
        cls,
        project_id: str,
        projects: Container[Project] | None = None
    ) -> Project:
        """Retrieve details on a Wistia project.

        The optional parameter `projects` contains info on all projects in
        the Wistia account, such as a response from the `Projects:list` API.

        :raises NoSuchProject: If the project doesn't exist on Wistia
        """
        if not projects:
            projects = WistiaDataApi.list_all_projects()

        for p in projects:
            if p.hashed_id == project_id:
                project = p
                break
        else:
            raise NoSuchProject(project_id)

        return project



from __future__ import annotations

from typing import TypeVar

from requests import HTTPError

from .api_base import _BaseWistiaApi
from .config import WistiaConfig
from .errors import *
from .log import LOG
from .models import *
from .utils.parse import resolve_contents


T = TypeVar('T')


class WistiaDataApi(_BaseWistiaApi):
    """
    Helper class to interact with the Wistia Data API (docs below)
      https://wistia.com/support/developers/data-api

    Fully implements the following sections in the API documentation:

        - Paging and Sorting Responses
        - Projects
        - Medias
        - Customizations
        - Captions

    """
    _API_ENDPOINT = WistiaConfig.API_URL

    # --------------------------
    # -       PROJECTS         -
    # --------------------------

    @classmethod
    def list_all_projects(
        cls,
        sort_by: SortBy | None = None,
        sort_dir: SortDir | None = None,
        per_page=_BaseWistiaApi._MAX_PER_PAGE
    ) -> Container[Project]:
        """
        Retrieve a list of Projects in the account, via the
        `Projects:list` API:
          https://wistia.com/support/developers/data-api#projects_list

        Defaults to sorting by Project ID. You can pass the`sort_by` argument
        to sort by another value.

        :raises NoSuchProject: If the project does not exist on Wistia
        """
        params = {}
        if sort_by:
            params['sort_by'] = sort_by.value
        if sort_dir:
            params['sort_direction'] = sort_dir.value

        try:
            data = cls.list_page(
                WistiaConfig.PROJECTS_URL,
                data_model=Project,
                per_page=per_page,
                params=params
            )
        except HTTPError:
            raise

        return data

    @classmethod
    def list_project(
        cls,
        project_id: str,
        sort_by: SortBy | None = None,
        sort_dir: SortDir | None = None,
        per_page=500,
        model_cls: type[T] = Media,
    ) -> Container[T | Media]:
        """
        Get all medias (generally videos) for a Wistia project, via the
        `Projects#show` API:
          https://wistia.com/support/developers/data-api#projects_show

        Defaults to sorting by Project ID. You can pass the`sort_by` argument
        to sort by another value.

        :raises NoSuchProject: If the project does not exist on Wistia
        """
        params = {'project_id': project_id}
        if sort_by:
            params['sort_by'] = sort_by.value
        if sort_dir:
            params['sort_direction'] = sort_dir.value

        try:
            data = cls.list_page(
                WistiaConfig.PROJECTS_SHOW_URL.format(project_id=project_id),
                data_model=model_cls,
                data_key='medias',
                per_page=per_page,
                params=params
            )
        except HTTPError as e:
            raise NoSuchProject(project_id) \
                if cls._has_resp_status(e, 404) else e

        return data

    @classmethod
    def create_project(
        cls,
        project_name: str | None = None,
        admin_email: str | None = None,
        public_upload: bool | None = None,
        public_download: bool | None = None,
        public: bool | None = None
    ) -> Project:
        """
        Create a new project in your account, via the `Projects:create` API:
          https://wistia.com/support/developers/data-api#projects_create

        """
        params = {'name': project_name}
        if admin_email:
            params['adminEmail'] = admin_email
        if public_upload is not None:
            params['anonymousCanUpload'] = 1 if public_upload else 0
        if public_download is not None:
            params['anonymousCanDownload'] = 1 if public_download else 0
        if public is not None:
            params['public'] = 1 if public else 0

        r = cls.session().post(
            WistiaConfig.PROJECTS_URL,
            params=params
        )
        r.raise_for_status()

        return Project.from_dict(r.json())

    @classmethod
    def update_project(
        cls, project_id: str,
        project_name: str | None = None,
        public_upload: bool | None = None,
        public_download: bool | None = None,
        public: bool | None = None
    ) -> Project:
        """
        Update an existing project in your account, via the
        `Projects:update` API:
          https://wistia.com/support/developers/data-api#projects_update

        """
        params = {}
        if project_name:
            params['name'] = project_name
        if public_upload is not None:
            params['anonymousCanUpload'] = 1 if public_upload else 0
        if public_download is not None:
            params['anonymousCanDownload'] = 1 if public_download else 0
        if public is not None:
            params['public'] = 1 if public else 0

        r = cls.session().put(
            WistiaConfig.PROJECTS_SHOW_URL.format(project_id=project_id),
            params=params
        )
        r.raise_for_status()

        return Project.from_dict(r.json())

    @classmethod
    def delete_project(cls, project_id: str):
        """
        Delete a project in your account, via the `Projects:delete` API:
          https://wistia.com/support/developers/data-api#projects_delete

        :return: A boolean indicating whether the video was successfully
          deleted.
        """
        return cls.handle_delete(
            WistiaConfig.PROJECTS_SHOW_URL.format(project_id=project_id)
        )

    @classmethod
    def copy_project(
        cls,
        project_id: str,
        admin_email: str | None = None
    ) -> Project:
        """
        Copies an existing project in your account, including all media and
        sections, via the `Projects:copy` API:
          https://wistia.com/support/developers/data-api#projects_copy

        """
        params = {}
        if admin_email:
            params['adminEmail'] = admin_email

        r = cls.session().post(
            WistiaConfig.PROJECTS_COPY_URL.format(project_id=project_id),
            params=params
        )
        r.raise_for_status()

        return Project.from_dict(r.json())

    # --------------------------
    # -        MEDIAS          -
    # --------------------------

    @classmethod
    def list_videos(
        cls,
        project_id: str | None = None,
        video_name: str | None = None,
        media_type: MediaType | None = MediaType.VIDEO,
        video_id: str | None = None,
        sort_by: SortBy | None = None,
        sort_dir: SortDir | None = None
    ) -> Container[Video]:
        """
        Get all videos for a Wistia project or by other criteria, via the
        `Medias#list` API:
          https://wistia.com/support/developers/data-api#medias_list

        Defaults to sorting by Project ID. You can pass the`sort_by` argument
        to sort by another value.

        :raises NoSuchProject: If the project does not exist on Wistia
        """
        return cls.list_medias(
            project_id=project_id,
            media_name=video_name,
            media_type=media_type,
            media_id=video_id,
            sort_by=sort_by,
            sort_dir=sort_dir,
            model_cls=Video
        )

    @classmethod
    def list_medias(
        cls,
        project_id: str | None = None,
        media_name: str | None = None,
        media_type: MediaType | None = None,
        media_id: str | None = None,
        sort_by: SortBy | None = None,
        sort_dir: SortDir | None = None,
        model_cls: type[T] = Media
    ) -> Container[T | Media]:
        """
        Get all medias for a Wistia project or by other criteria, via the
        `Medias#list` API:
          https://wistia.com/support/developers/data-api#medias_list

        Defaults to sorting by Project ID. You can pass the`sort_by` argument
        to sort by another value.

        :raises NoSuchProject: If the project does not exist on Wistia
        """
        params = {}
        if project_id:
            params['project_id'] = project_id
        if media_name:
            params['name'] = media_name
        if media_type:
            params['type'] = media_type.value
        if media_id:
            params['hashed_id'] = media_id
        if sort_by:
            params['sort_by'] = sort_by.value
        if sort_dir:
            params['sort_direction'] = sort_dir.value

        try:
            data = cls.list_page(
                WistiaConfig.MEDIAS_URL,
                data_model=model_cls,
                params=params
            )
        except HTTPError as e:
            raise NoSuchProject(project_id) \
                if cls._has_resp_status(e, 404) else e

        return data

    @classmethod
    def get_video(
        cls,
        video_id: str,
        model_cls: type[T] = Video
    ) -> T | Video:
        """
        Get information on a Wistia video, via the `Medias#show` API:
          https://wistia.com/support/developers/data-api#medias_show

        :raises NoSuchMedia: If the video does not exist on Wistia
        """
        r = cls.session().get(
            WistiaConfig.MEDIAS_SHOW_URL.format(media_id=video_id)
        )

        try:
            r.raise_for_status()
        except HTTPError as e:
            raise NoSuchMedia(video_id) if cls._has_resp_status(e, 404) else e

        return model_cls.from_dict(r.json())

    @classmethod
    def get_media(cls, media_id: str) -> Media:
        """
        Get information on a Wistia media, via the `Medias#show` API:
          https://wistia.com/support/developers/data-api#medias_show

        :raises NoSuchMedia: If the media does not exist on Wistia
        """
        return cls.get_video(media_id, model_cls=Media)

    @classmethod
    def update_video(
        cls,
        video_id: str,
        video_name: str | None = None,
        video_desc: str | None = None,
        thumbnail_media_id: str | None = None
    ) -> Video:
        """
        Updates attributes on a video, via the `Medias#update` API:
          https://wistia.com/support/developers/data-api#medias_update

        :param video_id: The hashed id of the video on Wistia.
        :param video_name: An optional new name for the video.
        :param video_desc: An optional new description for this video. Accepts
          plain text or markdown.
        :param thumbnail_media_id: The Wistia hashed ID of an image that will
          replace the still that’s displayed before the player starts playing.
        :raises NoSuchMedia: If the video does not exist on Wistia
        """
        return cls.update_media(
            media_id=video_id,
            media_name=video_name,
            media_desc=video_desc,
            thumbnail_media_id=thumbnail_media_id,
            model_cls=Video
        )

    @classmethod
    def update_media(
        cls,
        media_id: str,
        media_name: str | None = None,
        media_desc: str | None = None,
        thumbnail_media_id: str | None = None,
        model_cls: type[T] = Media
    ) -> T | Media:
        """
        Updates attributes on a media (generally a video), via the
        `Medias#update` API:
          https://wistia.com/support/developers/data-api#medias_update

        :param media_id: The hashed id of the media on Wistia.
        :param media_name: An optional new name for the media.
        :param media_desc: An optional new description for this media. Accepts
          plain text or markdown.
        :param thumbnail_media_id: The Wistia hashed ID of an image that will
          replace the still that’s displayed before the player starts playing.
        :param model_cls: The model class to load the response data into.
        :raises NoSuchMedia: If the media does not exist on Wistia
        """
        data = {}
        if media_name:
            data['name'] = media_name
        if media_desc:
            data['description'] = media_desc
        if thumbnail_media_id:
            data['new_still_media_id'] = thumbnail_media_id

        r = cls.session().put(
            WistiaConfig.MEDIAS_SHOW_URL.format(media_id=media_id),
            data=data
        )

        try:
            r.raise_for_status()
        except HTTPError as e:
            raise NoSuchMedia(media_id) if cls._has_resp_status(e, 404) else e

        return model_cls.from_dict(r.json())

    @classmethod
    def delete_media(cls, media_id: str):
        """
        Deletes a media (generally a video) from Wistia, via the
        `Medias#delete` API:
          https://wistia.com/support/developers/data-api#medias_delete

        :return: A boolean indicating whether the video was successfully
          deleted.
        """
        return cls.handle_delete(
            WistiaConfig.MEDIAS_SHOW_URL.format(media_id=media_id)
        )

    # Alias
    delete_video = delete_media

    @classmethod
    def copy_video(
        cls,
        video_id: str,
        dest_project_id: str | None = None,
        owner: str | None = None,
        model_cls: type[T] = Video
    ) -> T | Video:
        """
        Copy a video, optionally to another project, via the `Medias#copy` API:
          https://wistia.com/support/developers/data-api#medias_copy

        :raises NoSuchMedia: If the video does not exist on Wistia
        :raises NoSuchProject: If the project does not exist on Wistia
        """
        data = {}
        if dest_project_id:
            data['project_id'] = dest_project_id
        if owner:
            data['owner'] = owner

        r = cls.session().post(
            WistiaConfig.MEDIAS_COPY_URL.format(media_id=video_id),
            json=data
        )

        try:
            r.raise_for_status()
        except HTTPError as e:
            if cls._has_resp_status(e, 404):
                if dest_project_id and dest_project_id in r.text:
                    # Project does not exist
                    raise NoSuchProject(dest_project_id)
                # Video does not exist
                raise NoSuchMedia(video_id)
            else:
                raise e

        return model_cls.from_dict(r.json())

    @classmethod
    def copy_media(
        cls,
        media_id: str,
        dest_project_id: str | None = None,
        owner: str | None = None
    ) -> Media:
        """
        Copy a media, optionally to another project, via the `Medias#copy` API:
          https://wistia.com/support/developers/data-api#medias_copy

        :raises NoSuchMedia: If the media does not exist on Wistia
        :raises NoSuchProject: If the project does not exist on Wistia
        """
        return cls.copy_video(
            media_id,
            dest_project_id,
            owner,
            model_cls=Media
        )

    @classmethod
    def get_stats_for_video(cls, video_id: str) -> VideoStats:
        """
        Get aggregated tracking stats on a Wistia video, via the
        `Medias#stats` API:
          https://wistia.com/support/developers/data-api#medias_stats

        :raises NoSuchMedia: If the video does not exist on Wistia
        """
        r = cls.session().get(
            WistiaConfig.MEDIAS_STATS_URL.format(media_id=video_id))

        try:
            r.raise_for_status()
        except HTTPError as e:
            raise NoSuchMedia(video_id) if cls._has_resp_status(e, 404) else e

        return VideoStats.from_dict(r.json())

    # --------------------------
    # -     CUSTOMIZATIONS     -
    # --------------------------

    @classmethod
    def get_customizations(cls, video_id: str) -> Customizations:
        """
        Get customizations for a video on Wistia, via the
        `Customizations#show` API:
          https://wistia.com/support/developers/data-api#customizations_show

        :raises NoSuchMedia: If the video does not exist on Wistia
        """
        r = cls.session().get(
            WistiaConfig.CUSTOMIZATION_URL.format(media_id=video_id))

        try:
            r.raise_for_status()
        except HTTPError as e:
            raise NoSuchMedia(video_id) if cls._has_resp_status(e, 404) else e

        return Customizations.from_dict(r.json())

    @classmethod
    def create_customizations(
        cls,
        video_id: str,
        customizations: Customizations
    ) -> Customizations:
        """
        Overwrites the customizations for a video on Wistia, via the
        `Customizations#create` API:
          https://wistia.com/support/developers/data-api#customizations_create

        :return: The new customizations on the video
        :raises NoSuchMedia: If the video does not exist on Wistia
        """
        r = cls.session().post(
            WistiaConfig.CUSTOMIZATION_URL.format(media_id=video_id),
            json=customizations.to_dict()
        )

        try:
            r.raise_for_status()
        except HTTPError as e:
            raise NoSuchMedia(video_id) if cls._has_resp_status(e, 404) else e

        return Customizations.from_dict(r.json())

    @classmethod
    def update_customizations(
        cls,
        video_id: str,
        customizations: Customizations
    ) -> Customizations:
        """
        Updates the customizations for a video on Wistia, via the
        `Customizations#update` API:
          https://wistia.com/support/developers/data-api#customizations_update

        :return: The new customizations on the video
        :raises NoSuchMedia: If the video does not exist on Wistia
        """
        r = cls.session().put(
            WistiaConfig.CUSTOMIZATION_URL.format(media_id=video_id),
            json=customizations.to_dict()
        )

        try:
            r.raise_for_status()
        except HTTPError as e:
            raise NoSuchMedia(video_id) if cls._has_resp_status(e, 404) else e

        return Customizations.from_dict(r.json())

    @classmethod
    def delete_customizations(cls, video_id: str):
        """
        Deletes the customizations for a video on Wistia, via the
        `Customizations#delete` API:
          https://wistia.com/support/developers/data-api#customizations_delete

        """
        return cls.handle_delete(
            WistiaConfig.CUSTOMIZATION_URL.format(media_id=video_id))

    # --------------------------
    # -       CAPTIONS         -
    # --------------------------

    @classmethod
    def list_captions(cls, video_id: str) -> Container[VideoCaptions]:
        """
        Retrieves all the captions on a Wistia video, via the
        `Captions#index` API:
          https://wistia.com/support/developers/data-api#captions_index

        The text of the captions will be in SRT format.

        :raises NoSuchMedia: If the video does not exist on Wistia.
        """
        r = cls.session().get(
            WistiaConfig.ALL_CAPTIONS_URL.format(media_id=video_id))

        try:
            r.raise_for_status()
        except HTTPError as e:
            raise NoSuchMedia(video_id) if r.status_code == 404 else e

        list_of_captions = VideoCaptions.from_list(r.json())

        # Return data as a `Container` type, so it's easier for a user
        # to format (and visualize) the data as needed.
        return Container(VideoCaptions, list_of_captions)

    @classmethod
    def get_captions(
        cls,
        video_id: str,
        lang_code: LanguageCode
    ) -> VideoCaptions | None:
        """
        Retrieves the captions for a specific language on a Wistia video,
        via the `Captions#show` API:
          https://wistia.com/support/developers/data-api#captions_show

        The text of the captions will be in SRT format. If no captions exist
        for the specified language, a `None` value is returned.
        """
        r = cls.session().get(
            WistiaConfig.LANG_CAPTIONS_URL.format(
                media_id=video_id,
                lang_code=lang_code.value
            )
        )

        try:
            r.raise_for_status()
        except HTTPError:
            if r.status_code == 404:
                return None
            else:
                # Unexpected error
                raise

        return VideoCaptions.from_dict(r.json())

    @classmethod
    def create_captions(
        cls,
        video_id: str,
        lang_code: LanguageCode | None = None,
        srt_contents: str | None = None,
        srt_file: str | None = None
    ) -> None:
        """
        Create new captions for a given language on a Wistia video, via the
        `Captions#create` API:
          https://wistia.com/support/developers/data-api#captions_create

        Note that this always add new captions, regardless of whether captions
        already exist for the language. Please use :meth:`update_captions`
        if we need to replace captions for a language instead.

        :param video_id: The Wistia video to add the new captions for.
        :param lang_code: An optional ISO-639–2 language code for the file. If
          left unspecified, the language code will be detected automatically.
        :param srt_contents: The caption text in SRT format.
        :param srt_file: The path to an SRT file.
        :raises ContentIsEmpty: If one of `srt_contents` or `srt_file` is not
          provided.
        """
        data = {'caption_file': resolve_contents(srt_file, srt_contents)}
        if lang_code is not None:
            data['language'] = lang_code.value

        r = cls.session().post(
            WistiaConfig.ALL_CAPTIONS_URL.format(media_id=video_id),
            json=data
        )

        # check if the request was a success
        r.raise_for_status()

        # add an explicit return value
        return None

    @classmethod
    def update_captions(
        cls,
        video_id: str,
        lang_code: LanguageCode,
        srt_contents: str | None = None,
        srt_file: str | None = None
    ) -> None:
        """
        Replace captions for a given language on a Wistia video, via the
        `Captions#update` API:
          https://wistia.com/support/developers/data-api#captions_update

        **Note**: it is generally safer to call this method instead of
        :meth:`create_captions`, which always adds new captions for a given
        language; this method falls back to :meth:`create_captions` in case
        captions don't exist for the specified language.

        :param video_id: The Wistia video to add the new captions for.
        :param lang_code: An ISO-639–2 language code for the file. If
          left unspecified, the language code will be detected automatically.
        :param srt_contents: The caption text in SRT format.
        :param srt_file: The path to an SRT file.
        :raises ContentIsEmpty: If one of `srt_contents` or `srt_file` is not
          provided.
        """
        srt_contents = resolve_contents(srt_file, srt_contents)
        data = {'caption_file': srt_contents}

        r = cls.session().put(
            WistiaConfig.LANG_CAPTIONS_URL.format(
                media_id=video_id, lang_code=lang_code.value),
            json=data)

        # check if the request was a success
        try:
            r.raise_for_status()
        except HTTPError:
            if r.status_code != 404:
                raise
            # Captions don't exist for the specified language
            # Ref: https://wistia.com/support/developers/data-api#the-response-27  # noqa: E501
            LOG.info(
                '%s: No captions on video, will attempt to add them. '
                'lang_code=%s', video_id, lang_code.value)
            cls.create_captions(video_id, lang_code, srt_contents)

        # add an explicit return value
        return None

    @classmethod
    def delete_captions(
        cls,
        video_id: str,
        lang_code: LanguageCode
    ):
        """
        Deletes the captions for a video on Wistia, via the
        `Captions#delete` API:
          https://wistia.com/support/developers/data-api#captions_delete

        """
        return cls.handle_delete(
            WistiaConfig.LANG_CAPTIONS_URL.format(
                media_id=video_id,
                lang_code=lang_code.value
            )
        )

    @classmethod
    def order_captions(
        cls,
        video_id: str,
        automated: bool = False,
        rush: bool = True,
        automatically_enable: bool = True
    ) -> None:
        """
        Purchase English captions on a Wistia video, via the
        `Captions#purchase` API:
          https://wistia.com/support/developers/data-api#captions_purchase

        :param video_id: The hashed id of the video to order captions for.
        :param automated: Set to true to order computer-generated
          captions ($0.25/minute) or false to order human-generated
          captions ($2.50/minute).
          Defaults to false.
        :param rush: Set to true to enable a rush order (one business day
          turnaround, $4.00/minute), or false to retain the standard four
          business day turnaround for human-generated captions ($2.50/minute).
          Defaults to true.
        :param automatically_enable: Set to true to automatically enable
          captions for the video as soon as the order is ready or false to
          hold the captions for review before manually enabling them.
          Defaults to true.
        :raises VideoHasCaptions: If English captions already exist for
          the video.
        :raises NoSuchMedia: If the video does not exist.

        """
        params = {
            'automated': automated,
            'rush': rush,
            'automatically_enable': automatically_enable
        }

        r = cls.session().post(
            WistiaConfig.CAPTIONS_ORDER_URL.format(media_id=video_id),
            params=params
        )

        try:
            r.raise_for_status()
        except HTTPError:
            # Ref: https://wistia.com/support/developers/data-api#the-response-29  # noqa: E501
            if r.status_code == 400:
                raise VideoHasCaptions(video_id)
            elif r.status_code == 404:
                raise NoSuchMedia(video_id)
            else:
                raise

        return None



from json import loads

from wystia.utils.response import format_error


def test_format_error():
    msg, code, status = 'Test message', 'MyCode', '402'

    r = format_error(msg, code, status)

    assert 'body' in r
    assert 'statusCode' in r

    assert r['statusCode'] == int(status)

    body_dict = loads(r['body'])

    assert body_dict.get('success') is False
    assert 'error' in body_dict



"""
Common test fixtures and utilities, for unit tests
"""
import pytest
from pytest_mock import MockerFixture

from wystia import WistiaDataApi


@pytest.fixture
def mock_log(mocker: MockerFixture):
    return mocker.patch('wystia.errors.LOG')


@pytest.fixture(scope='session', autouse=True)
def global_setup():
    """Setup that runs before the unit test suite."""
    WistiaDataApi.configure(None)
    WistiaDataApi.reset_request_count()


@pytest.fixture(autouse=True)
def mock_request(mocker: MockerFixture):
    """
    Mock the underlying :meth:`request.Session.request` method
    so that any HTTP requests are mocked.
    """
    mocker.patch('wystia.api_base.SessionWithRetry.request')


@pytest.fixture
def mock_from_dict(mocker: MockerFixture):
    """
    Mock the underlying :meth:`JSONSerializable.from_dict` method
    so that the de-serialization process is mocked.
    """
    mocker.patch('dataclass_wizard.JSONSerializable.from_dict')


@pytest.fixture
def mock_open(mocker: MockerFixture):
    mocker.patch('wystia.api_upload.open', return_value=b'')



from __future__ import annotations

__all__ = ['as_bool',
           'as_int',
           'as_str',
           'as_list',
           'as_datetime']

from datetime import datetime
from numbers import Number
from typing import Any


def as_bool(o: str | bool, default=False):
    """
    Return `o` if already a boolean, otherwise return the boolean value for a
    string. If `o` is None or an empty string, return `default` instead.

    """
    if isinstance(o, bool):
        return o

    if not o:
        return default

    return o.upper() == 'TRUE'


def as_type(o: Any, _type: type = str, default=None, raise_=True):
    if isinstance(o, _type):
        return o

    if not o:
        return default

    try:
        return _type(o)
    except ValueError:
        if raise_:
            raise
        return default


def as_int(o: str | int, default=0, raise_=True):
    """
    Return `o` if already a int, otherwise return the int value for a
    string. If `o` is None or an empty string, return `default` instead.

    If `o` cannot be converted to an int, raise an error if `raise_` is true,
    other return `default` instead.

    """
    return as_type(o, int, default, raise_)


def as_str(o: str | None, default='', raise_=True):
    """
    Return `o` if already a str, otherwise return the string value for `o`.
    If `o` is None or an empty string, return `default` instead.

    If `o` cannot be converted to an str, raise an error if `raise_` is true,
    other return `default` instead.

    """
    return as_type(o, str, default, raise_) or default


def as_list(o: str | list[str], sep=','):
    """
    Return `o` if already a list. If `o` is None or an empty string,
    return an empty list. Otherwise, split the string on `sep` and
    return the list result.

    """
    if not o:
        return []

    if isinstance(o, list):
        return o

    return o.split(sep)


def as_datetime(o: str | Number | datetime, default=None, raise_=True):
    """
    Return `o` if already a :class:`datetime` object, otherwise convert the
    object to a :class:`datetime` object using the below logic.

        * ``str``: convert date strings (in ISO format) via the built-in
          ``fromisoformat`` method.
        * ``Number`` (int or float): Convert a numeric timestamp via the
            built-in `fromtimestamp`` method.

    If `o` is None or false-y, return `default` instead.

    Otherwise, if we're unable to convert the value of `o` to a
    :class:`datetime` as expected, raise an error if the `raise_` parameter
    is true.

    """
    if not o:
        return default

    if isinstance(o, datetime):
        return o

    try:
        if isinstance(o, Number):
            # noinspection PyTypeChecker
            return datetime.fromtimestamp(o)

        return datetime.fromisoformat(o)

    except (TypeError, ValueError):
        if raise_:
            raise
        return default



"""
Tests for the module `types`
"""
from datetime import datetime

import pytest

from wystia.utils.parse.types import *
from .conftest import does_not_raise


@pytest.mark.parametrize(
    'input,default,expected',
    [
        [None, False, False],
        ['tRuE', False, True],
        [True, False, True],
        [False, True, False],
        ['Nope', True, False],
    ]
)
def test_as_bool(input, default, expected):
    actual = as_bool(input, default)
    assert actual == expected


@pytest.mark.parametrize(
    'input,expected,default,raise_,expectation',
    [
        [None, 0, 0, True, does_not_raise()],
        [1, 1, 0, True, does_not_raise()],
        ['3', 3, 1, True, does_not_raise()],
        ['a', 3, 3, False, does_not_raise()],
        ['a', None, 3, True, pytest.raises(ValueError)],
        [172.75, 172, 3, True, does_not_raise()],
    ]
)
def test_as_int(input, expected, default, raise_, expectation):
    with expectation:
        actual = as_int(input, default, raise_)
        assert actual == expected


@pytest.mark.parametrize(
    'input,expected,default,raise_,expectation',
    [
        [None, '0', '0', True, does_not_raise()],
        [1, '1', 0, True, does_not_raise()],
        ['', 'hello world', 'hello world', True, does_not_raise()],
        ['testing', 'testing', 'hello world', True, does_not_raise()],
    ]
)
def test_as_str(input, expected, default, raise_, expectation):
    with expectation:
        actual = as_str(input, default, raise_)
        assert actual == expected


@pytest.mark.parametrize(
    'input,sep,expected',
    [
        [None, '|', []],
        ['True', '|', ['True']],
        ['hello,world', ',', ['hello', 'world']],
        ['hello|world', ',', ['hello|world']],
        ['hello|world', '|', ['hello', 'world']],
        ['this is a test', ' ', ['this', 'is', 'a', 'test']],
        [['this is a test'], ' ', ['this is a test']],
    ]
)
def test_as_list(input, sep, expected):
    actual = as_list(input, sep)
    assert actual == expected


@pytest.mark.parametrize(
    'input,expected,default,raise_,expectation',
    [
        [None, None, None, True, does_not_raise()],
        ['', datetime.min, datetime.min, True, does_not_raise()],
        [12345, datetime.fromtimestamp(12345), None, True, does_not_raise()],
        [datetime(2001, 11, 22, 23, 59, 1), datetime(2001, 11, 22, 23, 59, 1),
         None, True, does_not_raise()],
        ['2021-01-02T03:00:01', datetime(2021, 1, 2, 3, 0, 1),
         None, True, does_not_raise()],
        ['abc', 'testing', None, True, pytest.raises(ValueError)],
        ['abc', 'testing', 'testing', False, does_not_raise()],
    ]
)
def test_as_datetime(input, expected, default, raise_, expectation):
    with expectation:
        actual = as_datetime(input, default, raise_)
        assert actual == expected



""""
Base API Classes
"""
from __future__ import annotations

import functools
from abc import ABC, abstractmethod
from typing import Any

from dataclass_wizard.abstractions import W
from requests import Session, RequestException

from .constants import WISTIA_API_TOKEN
from .log import LOG
from .models import Container
from .requests_models import SessionWithRetry, prefix_url_session


class _BaseApi(ABC):
    """
    Abstract base class for sending requests to an :attr:`API_ENDPOINT`
    """

    # Base API endpoint
    _API_ENDPOINT: str = NotImplemented

    # Requests Session object, can be used if session caching is needed
    _SESSION: Session | None = None

    @classmethod
    def _get(cls, api, **kwargs):
        """Shortcut to make a ``GET`` request via the API"""
        return cls.request('GET', api, **kwargs)

    @classmethod
    def request(cls, method, api, **kwargs) -> dict | list:
        """
        Makes an API request using an HTTP `method`.

        Returns the response data as a Python object, after validating that
        the request was a success.

        """
        # Make the API request
        r = cls._get_session().request(method, api, **kwargs)
        # Check if the request was a success
        r.raise_for_status()
        # Return the JSON response
        return r.json()

    @classmethod
    def session(cls):
        """Shortcut to retrieve the :class:`requests.Session` object"""
        return cls._get_session()

    @classmethod
    @abstractmethod
    def _get_session(cls) -> Session:
        """
        Return the :class:`requests.Session` object for an API request.
        """

    @staticmethod
    def _has_resp_status(e: RequestException, status: int) -> bool:
        """Check if the response in an error has a specified status code"""
        return e.response is not None and e.response.status_code == status


class _BaseWistiaApi(_BaseApi):
    """
    Base class for sending requests to the Wistia API.

    Note: any sub-classes will still need to override the :attr:`API_ENDPOINT`

    """
    # Default to the value specified via the environment
    _API_TOKEN: str = WISTIA_API_TOKEN

    # Maximum results per page; also documented in the link below
    #   https://wistia.com/support/developers/data-api#paging
    _MAX_PER_PAGE: int = 100

    # This attribute keeps a running count of the total API requests made.
    #
    # The current rate limit for the Wistia API is 600 requests / min as
    # documented in the link below. Use `request_count()` to retrieve the
    # number of API requests made, and any custom application logic to sleep
    # when the request limit is reached. Don't forget to reset the running
    # total via `reset_request_count()`.
    #   https://wistia.com/support/developers/data-api#rate
    _REQUEST_COUNT: int = 0

    @staticmethod
    def configure(api_token: str):
        """
        Sets the API token used to authenticate requests to the Wistia API.
        """
        _BaseWistiaApi._API_TOKEN = api_token

    @classmethod
    def request_count(cls):
        """Return the running count of API requests to the Wistia API."""
        return cls._REQUEST_COUNT

    @classmethod
    def reset_request_count(cls):
        """
        Reset (clear) the running count of API requests to the Wistia API.
        """
        _BaseWistiaApi._REQUEST_COUNT = 0

    @classmethod
    def session(
        cls,
        additional_status_force_list: list[int] | None = None
    ) -> Session:
        """
        Return a new :class:`requests.Session` object.

        `additional_status_force_list` is an optional list of additional HTTP
        response status codes to retry on, in additional to the
        `DEFAULT_STATUS_FORCE_LIST`.

        """
        return cls._get_session(additional_status_force_list)

    @classmethod
    def list_page(
        cls,
        url: str,
        data_model: type[W] = None,
        data_key: str | None = None,
        per_page: int = _MAX_PER_PAGE,
        **kwargs
    ) -> Container[W] | list[dict]:
        """
        Makes an HTTP GET request to the Wistia API. If we get back exactly
        ``max_per_page`` results, this indicates there are more results
        available, so we will pass the ``page`` parameter to retrieve all the
        results for an API request.

        If `data_model` is passed as a valid `JSONWizard` sub-class, the JSON
        response will be de-serialized into a :class:`Container` (a list) of
        instances. Otherwise, we return a list of `dict` objects instead.

        Used primarily for `list` methods in the Wistia API, as described
        below:
          https://wistia.com/support/developers/data-api#paging_and_sorting_responses

        :raises HTTPError: Raised for any 4xx or 5xx errors.
        :raises ConnectionError: Raised for any request timeouts or connection
          errors.
        """
        data = []
        page = 1

        page_data = cls._request_page(url, per_page=per_page, **kwargs)
        if data_key:
            page_data = page_data[data_key]
        data.extend(page_data)

        while len(page_data) == per_page:
            page += 1
            page_data = cls._request_page(url, page, per_page, **kwargs)
            if data_key:
                page_data = page_data[data_key]
            data.extend(page_data)

        if data_model:
            instances: list[W] = data_model.from_list(data)
            return Container(data_model, instances)

        return data

    @classmethod
    def handle_delete(cls, url: str) -> bool:
        """
        Makes an HTTP DELETE request to the Wistia API. If the response is a
        200 (OK) status code, this indicates the DELETE request was a success;
        otherwise, we log an error as it might be useful for debugging
        purposes.

        :param url: URL for the endpoint, without the base API prefix.
        :return: A boolean indicating whether the request was a success.
        """
        r = cls.session().delete(url)

        success = r.status_code == 200
        if not success:
            import inspect

            caller_name: str = inspect.stack()[1][3]
            api_name = caller_name.replace('_', ' ').title()

            LOG.error('Wistia %s API. status=%d, reason=%r, text=%s',
                      api_name, r.status_code, r.reason, r.text)

        return success

    @classmethod
    def _increment_count_decorator(cls, func):
        """
        Returns a decorator which wraps a function `func` to increment the
        :attr:`_REQUEST_COUNT` attribute on each function call.
        """
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _BaseWistiaApi._REQUEST_COUNT += 1
            return func(*args, **kwargs)

        return wrapper

    @classmethod
    def _get_session(cls, additional_status_force_list=None) -> Session:
        """
        Return the :class:`requests.Session` object for an API request.
        """
        session = cls._create_session(additional_status_force_list)
        session.request = cls._increment_count_decorator(session.request)

        return session

    @classmethod
    def _create_session(cls, additional_status_force_list=None):
        """
        Create and return a new :class:`request.Session` object, using `HTTP
        Basic` authentication as described here:
          https://wistia.com/support/developers/data-api#authentication

        In this implementation, the Session object automatically retries
        for common server-side error codes, and prefixes the url in all API
        requests with the :attr:`API_ENDPOINT` attribute.
        """
        session = SessionWithRetry(
            auth=('api', cls._API_TOKEN),
            additional_status_force_list=additional_status_force_list)

        return prefix_url_session(cls._API_ENDPOINT, session)

    @classmethod
    def _request_page(
            cls,
            url: str,
            page: int | None = None,
            per_page: int | None = None,
            **kwargs
    ) -> dict[str, Any]:
        """
        Requests a single page from the the Wistia API.
        """
        params = (kwargs.pop('params', None) or {}).copy()
        if page:
            params['page'] = page
        if per_page:
            params['per_page'] = per_page

        r = cls._get_session().request('GET', url, params=params, **kwargs)
        r.raise_for_status()

        return r.json()



from __future__ import annotations

__all__ = ['LanguageCode',
           'MediaType',
           'SortBy',
           'SortDir',
           'MediaStatus',
           'Container',
           'Project',
           'Media',
           'Video',
           'VideoStats',
           # Video Embed Options
           'Customizations',
           'UnalteredStillImageAsset',
           'Plugin',
           'ThumbnailTextOverlay',
           'ThumbnailTextOverlayV2',
           'VideoThumbnail',
           'MidrollLinkV1',
           'Link',
           'PostRollV1',
           'Style',
           'CaptionsV1',
           'Chapters',
           'ChapterList',
           'Eventbrite',
           'Share',
           'RequireEmailV1',
           'PasswordProtectedVideo',
           'Private',
           'Encrypted',
           # Video Captions
           'VideoCaptions',
           # Video Embed Data
           'VideoEmbedData',
           # Upload API models
           'UploadResponse']

import json
import pprint
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Generic, Iterator, Iterable, Any

from dataclass_wizard import JSONWizard, json_field
from dataclass_wizard.abstractions import W
from dataclass_wizard.type_def import Encoder

from .constants import RAISE_ON_UNKNOWN_KEY
from .log import LOG
from .utils.decorators import cached_property
from .utils.metaclasses import display_with_pformat
from .utils.parse import get_srt_duration


class LanguageCode(Enum):
    """
    The 3-character Language Codes for Wistia, as specified by ISO-639-2.
    """
    CHINESE = 'chi'
    ENGLISH = 'eng'
    FRENCH = 'fre'
    GERMAN = 'ger'
    ITALIAN = 'ita'
    SPANISH = 'spa'
    JAPANESE = 'jpn'


class MediaType(Enum):
    """
    Wistia Media Types, as documented in the link below:
      https://wistia.com/support/developers/data-api#medias_response

    """
    VIDEO = 'Video'
    AUDIO = 'Audio'
    IMAGE = 'Image'
    PDF = 'PdfDocument'
    WORD = 'MicrosoftOfficeDocument'
    SWF = 'Swf'
    UNKNOWN = 'UnknownType'


class SortBy(Enum):
    """
    The name of the field to sort by. Defaults to sorting by Project ID.
      https://wistia.com/support/developers/data-api#sorting

    """
    NAME = 'name'
    CREATED = 'created'
    UPDATED = 'updated'


class SortDir(Enum):
    """
    Specifies the direction of the sort, defaults to ASC (ascending) order.
      https://wistia.com/support/developers/data-api#sorting

    """
    DESC = 0
    ASC = 1


class MediaStatus(Enum):
    """Describes the current status of Media files.

    The status indicates which stage in processing the file is at.
      https://wistia.com/support/developers/data-api#media-status

    """
    # the file has been fully processed and is ready for embedding and viewing
    READY = 'ready'
    # the file is actively being processed
    PROCESSING = 'processing'
    # the file is waiting in the queue to be processed
    QUEUED = 'queued'
    # the file was unable to be processed (usually a format or size error)
    FAILED = 'failed'
    # custom enum member, indicates that media was taken down from Wistia
    # note: this is not part of the Wistia API
    NOT_FOUND = 'not_found'


class Container(list, Generic[W]):
    """
    List wrapper around a list of :class:`JSONWizard` (or a sub-class)
    instances.
    """
    __slots__ = ('_model', )

    def __init__(self, data_model: type[W], seq: Iterable[W] = ()):
        super().__init__(seq)
        self._model = data_model

    def __iter__(self) -> Iterator[W]:
        return super().__iter__()

    def __str__(self):
        """Control the value displayed when print(self) is called."""
        return pprint.pformat(self)

    def to_json(self, encoder: Encoder = json.dumps,
                ensure_ascii=False,
                **encoder_kwargs) -> str:
        """Convert the list of instances to a JSON string."""
        # noinspection PyArgumentList
        return self._model.list_to_json(
            self,
            encoder=encoder,
            ensure_ascii=ensure_ascii,
            **encoder_kwargs
        )

    def prettify(self, encoder: Encoder = json.dumps,
                 ensure_ascii=False,
                 **encoder_kwargs) -> str:
        """Convert the list of instances to a *prettified* JSON string."""
        return self.to_json(
            indent=2,
            encoder=encoder,
            ensure_ascii=ensure_ascii,
            **encoder_kwargs
        )


#########################
#   Data API - Models   #
#########################

@dataclass
class Project(JSONWizard, metaclass=display_with_pformat):
    """
    Project dataclass
    """
    class _(JSONWizard.Meta):
        raise_on_unknown_json_key = RAISE_ON_UNKNOWN_KEY

    hashed_id: str
    id: int
    name: str
    media_count: int
    created: datetime
    updated: datetime
    anonymous_can_upload: bool
    anonymous_can_download: bool
    public: bool
    public_id: str
    description: str | None


@dataclass
class Media(JSONWizard, metaclass=display_with_pformat):
    """
    Media dataclass

    """
    class _(JSONWizard.Meta):
        raise_on_unknown_json_key = RAISE_ON_UNKNOWN_KEY
        skip_defaults = True

    hashed_id: str
    id: int
    name: str
    type: MediaType
    created: datetime
    updated: datetime
    # Note: only videos have this attribute set; thumbnails and other
    # medias don't.
    duration: float | None = None
    status: MediaStatus = MediaStatus.NOT_FOUND
    description: str = ''
    progress: float = 0.0
    thumbnail: Thumbnail = None
    project: ProjectInfo | None = None
    embed_code: str = json_field('', repr=False, dump=False, default=None)
    assets: list[Asset] = None
    section: str | None = None

    def __post_init__(self):
        # Media (video) titles seem to include encoded characters
        # like '&amp;' in the response, but not in the title displayed
        # on the Wistia project page.
        self.name = self.name.replace('&amp;', '&')

        # Sometimes the description for media (videos) will come as a newline
        # in the response, but it's actually empty on the Wistia page. So
        # we'll replace a newline with an empty string; that way it's also
        # dropped when the `to_json` method is called, for example.
        if self.description == '\n':
            self.description = ''

    @cached_property
    def project_id(self) -> str:
        """Return the project's hashed id."""
        return self.project.hashed_id if self.project else ''

    def set_project_id(self, project_id: str):
        """Set the project's hashed id for the media."""
        if self.project:
            self.project.hashed_id = project_id
        else:
            self.project = ProjectInfo(project_id)


@dataclass
class Thumbnail:
    """
    Thumbnail dataclass

    """
    url: str
    width: int
    height: int


@dataclass
class Asset:
    """
    Asset dataclass

    """
    url: str
    file_size: int
    content_type: str
    type: str
    width: int | None = None
    height: int | None = None


@dataclass
class Video(Media, JSONWizard, metaclass=display_with_pformat):
    """
    Video dataclass

    """
    class _(JSONWizard.Meta):
        raise_on_unknown_json_key = RAISE_ON_UNKNOWN_KEY
        skip_defaults = True

    # Override the type annotations as needed.
    duration: float = 0.0
    project: ProjectInfo = None

    # Not included in GET '/v1/medias' response, but technically
    # still part of video metadata.
    has_audio_description: bool | None = None
    captions_enabled: bool | None = None
    overlay_text: str | None = None
    caption_duration: float | None = None
    num_captions: int | None = None
    ad_disabled: bool | None = None

    def __post_init__(self):
        if self.type is not MediaType.VIDEO:
            return

        # Check if `assets` are populated for the video.
        #
        # For some cases, like in the response for the `Projects#show` API,
        # neither the `assets` or `project` data will be populated.
        if self.assets is not None:
            for asset in self.assets:
                if asset.type == 'AlternateAudioFile':
                    self.has_audio_description = True
                    break
            else:
                self.has_audio_description = False

        if not self.duration:
            # There are rare cases when 'duration' field is missing from
            # response. This usually also means video is inaccessible from the
            # webpage, so we might need to contact Wistia Support to resolve
            # the issue.
            self.status = MediaStatus.FAILED
            self.duration = 0.0
            LOG.error(
                f'Video ({self.hashed_id}) is missing a required field '
                'in get_video response')

    @classmethod
    def load_video(cls, video_id: str) -> Video:
        """
        Retrieve video data from Wistia and return a new :class:`Video`
        object.
        """
        from .api_data import WistiaDataApi

        return WistiaDataApi.get_video(video_id)

    @classmethod
    def list_for_project(cls, project_id: str) -> list[Video]:
        """
        List videos for a Wistia project.
        """
        from .api_data import WistiaDataApi

        videos: Container[Video] = WistiaDataApi.list_project(
            project_id,
            model_cls=Video
        )

        for v in videos:
            v.project = ProjectInfo(project_id)

        return videos

    @property
    def has_captions(self) -> bool:
        """
        Indicates whether the Wistia video has a captions file.
        """
        return self.caption_duration is not None

    def process_captions(self, captions: Iterable[VideoCaptions]):
        """
        Process the response from the `Captions: Index` API for the video.

        Sets the following attributes on the object:
            - `num_captions`
            - `caption_duration`

        Ref: https://wistia.com/support/developers/data-api#captions_index
        """
        if not captions:
            return

        caption_durations: list[float] = []
        for caption in captions:
            # lc = LanguageCode(caption['language'])
            captions_end_seconds = get_srt_duration(caption.text)

            caption_durations.append(captions_end_seconds)

        self.num_captions = len(caption_durations)
        self.caption_duration = caption_durations[0]

    def process_customizations(self, customizations: Customizations):
        """
        Process the response from the `Customizations: Show` API for the video.

        Sets the following attributes on the object:
            - `ad_disabled`
            - `captions_enabled`
            - `overlay_text`

        Ref: https://wistia.com/support/developers/data-api#customizations_show
        """

        # Check if ad is explicitly disabled
        ad_disabled = customizations.audio_description_is_required is False
        overlay_text = (customizations.plugin.thumbnail_text_overlay_v2
                        or ThumbnailTextOverlayV2()).text

        self.ad_disabled = ad_disabled
        self.captions_enabled = customizations.plugin.captions_v1.on is True
        self.overlay_text = overlay_text.strip()


@dataclass
class ProjectInfo:
    """
    Project dataclass

    """
    hashed_id: str
    id: int = None
    name: str = None


@dataclass
class VideoStats(JSONWizard, metaclass=display_with_pformat):
    """
    VideoStats dataclass

    """
    class _(JSONWizard.Meta):
        raise_on_unknown_json_key = RAISE_ON_UNKNOWN_KEY

    id: int
    hashed_id: str
    name: str
    stats: Stats


@dataclass
class Stats(metaclass=display_with_pformat):
    """
    Stats dataclass

    """
    page_loads: int
    visitors: int
    percent_of_visitors_clicking_play: int
    plays: int
    average_percent_watched: int


############################
#   Video Customizations   #
############################

@dataclass
class Customizations(JSONWizard, metaclass=display_with_pformat):
    """
    Model class which represents customization data for a Wistia video - also
    known as "Embed Options" in the Wistia docs.

    Ref: https://wistia.com/support/developers/embed-options
    """
    class _(JSONWizard.Meta):
        raise_on_unknown_json_key = RAISE_ON_UNKNOWN_KEY
        skip_defaults = True

    player_color: str = ''
    still_url: str | None = None
    unaltered_still_image_asset: UnalteredStillImageAsset | None = None
    thumbnail_alt_text: str | None = None
    auto_play: bool | None = None
    silent_auto_play: bool | None = None
    end_video_behavior: str | None = None
    controls_visible_on_load: bool = True
    play_button: bool | None = None
    bpb_time: bool | None = None
    small_play_button: bool | None = None
    # noinspection SpellCheckingInspection
    play_bar: bool | None = json_field('playbar', all=True, default=None)
    volume_control: bool | None = None
    settings_control: bool | None = None
    playback_rate_control: bool | None = None
    quality_control: bool | None = None
    fullscreen_button: bool | None = None
    plugin: Plugin | None = None
    private: Private | None = None
    encrypted: Encrypted | None = None
    audio_description_is_required: bool | None = None
    # noinspection SpellCheckingInspection
    plays_inline: bool | None = json_field('playsinline', all=True, default=None)
    branding: bool | None = None
    show_customer_logo: bool | None = None
    customer_logo_image_url: str = ''
    customer_logo_target_url: str = ''
    chapters_on: bool | None = None
    spherical: bool | None = None
    vulcan: bool | None = None
    video_quality: str = ''
    muted: bool | None = None


@dataclass
class UnalteredStillImageAsset:
    """
    UnalteredStillImageAsset dataclass

    """
    url: str
    width: int | None = None
    height: int | None = None


@dataclass
class CaptionsV1:
    """
    CaptionsV1 dataclass

    """
    on: bool | None = None
    on_by_default: bool | None = None
    is_async: bool | None = json_field('async', all=True, default=None)
    language: str = ''


@dataclass
class Plugin:
    """
    Plugin dataclass

    """
    thumbnail_text_overlay_v2: ThumbnailTextOverlayV2 | None = json_field(
        'thumbnailTextOverlay-v2', all=True, default=None)
    # Possibly deprecated in favor of V2; best to avoid using it directly.
    thumbnail_text_overlay: ThumbnailTextOverlay | None = None
    video_thumbnail: VideoThumbnail | None = None
    midroll_link_v1: MidrollLinkV1 | None = json_field(
        'midrollLink-v1', all=True, default=None)
    post_roll_v1: PostRollV1 | None = json_field(
        'postRoll-v1', all=True, default=None)
    captions_v1: CaptionsV1 = json_field(
        'captions-v1', all=True, default_factory=CaptionsV1)
    chapters: Chapters | None = None
    eventbrite: Eventbrite | None = None
    share: Share | None = None
    require_email_v1: RequireEmailV1 | None = json_field(
        'requireEmail-v1', all=True, default=None)
    password_protected_video: PasswordProtectedVideo | None = None


@dataclass
class ThumbnailTextOverlayV2:
    """
    ThumbnailTextOverlayV2 dataclass

    """
    on: bool | None = None
    text: str = ''
    include_text_overlay: bool | None = None
    text_overlay_image_url: str | None = None
    is_async: bool | None = json_field('async', all=True, default=None)


@dataclass
class ThumbnailTextOverlay:
    """
    ThumbnailTextOverlay dataclass

    """
    on: bool | None = None
    text: str = ''
    include_text_overlay: bool | None = None
    chapters: Chapters | None = None
    captions_v1: CaptionsV1 | None = json_field(
        'captions-v1', all=True, default=None)
    is_async: bool | None = json_field('async', all=True, default=None)


@dataclass
class VideoThumbnail:
    """
    VideoThumbnail dataclass

    """
    on: bool
    hashed_id: str
    trim_start: int
    trim_end: int
    is_async: bool | None = json_field('async', all=True, default=None)


@dataclass
class MidrollLinkV1:
    """
    MidrollLinkV1 dataclass

    """
    on: bool | None = None
    links: list[Link] | None = None


@dataclass
class Link:
    """
    Link dataclass

    """
    name: str
    time: float
    duration: float
    text: str
    url: str
    position: str | None = None
    theme: str | None = None
    conversion_opportunity_id: int | None = None
    conversion_opportunity_key: str | None = None


@dataclass
class PostRollV1:
    """
    PostRollV1 dataclass

    """
    on: bool | None = None
    rewatch: bool | None = None
    text: str | None = None
    link: str | None = None
    time: float | None = None
    auto_size: bool | None = None
    cta_type: str | None = None
    style: Style | None = None
    alt_text: str | None = None
    raw: str | None = None
    image: str | None = None
    conversion_opportunity_key: str | None = None


@dataclass
class Style:
    """
    Style dataclass

    """
    background_color: str


@dataclass
class Chapters:
    """
    Chapters dataclass

    """
    on: bool | None = None
    visible_on_load: bool | None = None
    chapter_list: list[ChapterList] | None = None


@dataclass
class ChapterList:
    """
    ChapterList dataclass

    """
    id: int
    title: str
    time: float
    deleted: bool


@dataclass
class Eventbrite:
    """
    Eventbrite dataclass

    """
    on: bool | None = None
    event_id: str = ''
    text: str = ''
    time: float | None = None
    duration: int | None = None
    type: str | None = None
    conversion_opportunity_key: str | None = None


@dataclass
class Share:
    """
    Share dataclass

    """
    on: bool | None = None
    channels: str | None = None
    page_title: str | None = None
    page_url: str | None = None
    tweet_text: str | None = None
    override_url: bool | None = None
    conversion_opportunity_key: str | None = None
    download_type: str | None = None


@dataclass
class RequireEmailV1:
    """
    RequireEmailV1 dataclass

    """
    on: bool | None = None
    top_text: str | None = None
    bottom_text: str | None = None
    time: float | None = None
    ask_name: bool | None = None
    allow_skip: bool | None = None
    persistent_turnstile: bool | None = None
    conversion_opportunity_key: str | None = None
    submit_button_text: str | None = None
    is_async: bool | None = json_field('async', all=True, default=None)


@dataclass
class PasswordProtectedVideo:
    """
    PasswordProtectedVideo dataclass

    """
    on: bool | None = None
    challenge: str | None = None


@dataclass
class Private:
    """
    Private dataclass

    """
    show_comments: bool | None = None
    password_protect_on: bool | None = None


@dataclass
class Encrypted:
    """
    Encrypted dataclass

    """
    password_protect_password: str = ''


######################
#   Video Captions   #
######################

@dataclass
class VideoCaptions(JSONWizard, metaclass=display_with_pformat):
    """
    Video Captions dataclass

    """
    class _(JSONWizard.Meta):
        raise_on_unknown_json_key = RAISE_ON_UNKNOWN_KEY

    language: LanguageCode
    text: str
    english_name: str
    native_name: str
    is_draft: bool

    @cached_property
    def approx_total_duration(self) -> float:
        """Return the approximate duration (in seconds) of the captions.

        This duration is estimated based on the end timestamp in the SRT
        contents of the captions.
        """
        return get_srt_duration(self.text)


###########################
#   Upload API - Models   #
###########################

@dataclass
class UploadResponse(JSONWizard, metaclass=display_with_pformat):
    """
    Represents a response from the Wistia Upload API

    Example response format:
      https://wistia.com/support/developers/upload-api#response-format
    """
    class _(JSONWizard.Meta):
        raise_on_unknown_json_key = RAISE_ON_UNKNOWN_KEY

    hashed_id: str
    id: int
    name: str
    type: MediaType
    description: str | None
    account_id: int
    created: datetime
    updated: datetime
    progress: float
    thumbnail: Thumbnail
    duration: float | None = None
    status: MediaStatus = MediaStatus.QUEUED


############################
#   Media Embed - Models   #
############################

@dataclass
class VideoEmbedData(JSONWizard, metaclass=display_with_pformat):
    """
    VideoEmbedData dataclass

    """
    class _(JSONWizard.Meta):
        raise_on_unknown_json_key = RAISE_ON_UNKNOWN_KEY
        skip_defaults = True

    hashed_id: str
    name: str
    created_at: datetime
    duration: float
    assets: list[EmbedAsset]
    project_id: int
    stats: EmbedStats
    distillery_url: str
    account_key: str
    media_key: str
    type: str
    media_type: str
    progress: float
    status: int
    branding: bool
    enable_customer_logo: bool
    seo_description: str
    preload_preference: Any
    flash_player_url: str
    show_about: bool
    first_embed_for_account: bool
    first_share_for_account: bool
    keyframe_align: bool
    use_media_data_host_logic: bool
    tracking_transmit_interval: int
    # Annotating this field as a generic `dict` type for now, because
    # I've not seen this feature used before.
    integrations: dict
    # integrations: Integrations
    hls_enabled: bool
    embed_options: Customizations
    captions: list[EmbedCaption] = field(default_factory=list)
    transcript: Transcript | None = None

    # Not included in GET '/v1/medias' response, but technically
    # still part of video metadata.
    source_url: str = ''
    ad_url: str | None = None
    has_audio_description: bool = False

    def __post_init__(self):
        for asset in self.assets:
            if asset.type == 'original':
                self.source_url = asset.url.replace('.bin', '/file.mp4', 1)
            if asset.type == 'alternate_audio':
                self.has_audio_description = True
                self.ad_url = asset.url.replace('.bin', '/file.mp3', 1)

    @cached_property
    def captions_enabled(self) -> bool:
        return self.embed_options.plugin.captions_v1.on is True

    @cached_property
    def ad_enabled(self) -> bool:
        return self.embed_options.audio_description_is_required

    @cached_property
    def num_captions(self) -> int:
        return len(self.captions)

    @classmethod
    def load_video(cls, video_id: str) -> VideoEmbedData:
        """
        Retrieve video embed data from Wistia and return a new
        :class:`VideoEmbedData` object.

        """
        from .api_embed import WistiaEmbedApi

        return WistiaEmbedApi.get_data(video_id)


@dataclass
class Transcript:
    video_id: str


@dataclass
class Details:
    """
    Details dataclass

    """
    audio_description: bool | None = None
    language_metadata: LanguageMetadata | None = None


@dataclass
class EmbedAsset:
    """
    EmbedAsset dataclass

    """
    type: str
    slug: str
    display_name: str
    bitrate: int
    public: bool
    status: int
    progress: float
    url: str
    created_at: datetime
    details: Details = field(default_factory=Details)
    size: int | None = None
    ext: str = ''
    metadata: Metadata | None = None
    container: str | None = None
    codec: str | None = None
    segment_duration: int | None = None
    opt_vbitrate: int | None = None
    width: int | None = None
    height: int | None = None


@dataclass
class LanguageMetadata:
    name: str
    native_name: str
    right_to_left: bool


@dataclass
class Metadata:
    """
    Metadata dataclass

    """
    aspect_ratio: float | None = None
    av_stream_metadata: str | None = None
    average_bitrate: int | None = None
    early_max_bitrate: int | None = None
    frame_width: int | None = None
    frame_height: int | None = None
    frame_count: int | None = None
    max_bitrate: int | None = None
    served_by_media_api: int | None = None


@dataclass
class EmbedStats:
    """
    Stats dataclass

    """
    load_count: int
    play_count: int
    unique_load_count: int
    unique_play_count: int
    average_engagement: float


@dataclass
class Integrations:
    """
    Integrations dataclass

    TODO we don't use this feature, so no idea what goes here
    """
    pass


@dataclass
class EmbedCaption:
    """
    EmbedCaption dataclass

    """
    language: LanguageCode
    text: str



from __future__ import annotations

__all__ = ['resolve_contents']

from ...errors import ContentIsEmpty


def resolve_contents(
    file_path: str | None = None,
    contents: str | None = None,
    raise_=True
) -> str:
    """
    Resolves file contents, given two optional parameters.

    :param file_path: An optional path to the file to read.
    :param contents: The optional contents of a file.
    :param raise_: Whether to raise an error if the file contents are empty.
    :return: The resolved file contents.
    :raises ContentIsEmpty: If both `file_path` and `contents` are empty, and
      the `raise_` flag is enabled.
    """
    if file_path:
        with open(file_path, 'r') as f:
            contents = f.read()

    if not contents and raise_:
        raise ContentIsEmpty()

    return contents



"""
Utilities for API Gateway response formatting
"""
import json


def format_response(data, status=200):
    return {
        'body': json.dumps(data),
        'headers': {
            'Content-Type': 'application/json'
        },
        'statusCode': int(status)
    }


def format_error(msg, code='BadRequest', status=400):
    data = {
        'success': False,
        'error': {
            'code': code,
            'message': msg
        }
    }

    return format_response(data, status)



__all__ = [
    'does_not_raise'
]

import sys


# Ref: https://docs.pytest.org/en/6.2.x/example/parametrize.html#parametrizing-conditional-raising  # noqa: E501
if sys.version_info >= (3, 7):
    from contextlib import nullcontext as does_not_raise
else:
    from contextlib import ExitStack as does_not_raise



__all__ = ['total_seconds',
           'total_ms',
           'get_srt_duration']

from functools import reduce


def total_seconds(ts: str) -> str:
    """
    Converts a timestamp containing hours, minutes, seconds, and milliseconds
    (for example, in the "HH:mm:ss,SSS" format) to a string representing the
    total seconds, along with the millisecond part.

    For example, a string like "1:20:32,5" will be returned as "4832.005"

    Supports parsing the following input formats:
        - (H)H:mm:ss,SSS
        - (H)H:mm:ss.SSS
        - (H)H:mm:ss:SSS

    A modified version of the following (great) solution:
    https://stackoverflow.com/a/57610198

    """
    seconds, milliseconds = divmod(total_ms(ts), 1000)

    return f'{seconds}.{milliseconds:0>3}'


def total_ms(ts: str) -> int:
    """
    Converts a timestamp containing hours, minutes, seconds, and milliseconds
    (for example, in the "HH:mm:ss,SSS" format) to an integer value
    representing the total milliseconds.

    For example, a string like "1:20:32,5" will be returned as 4832005

    Supports parsing the following input formats:
        - (H)H:mm:ss,SSS
        - (H)H:mm:ss.SSS
        - (H)H:mm:ss:SSS

    A modified version of the following (great) solution:
    https://stackoverflow.com/a/57610198

    """
    try:
        h_m_s, milliseconds = ts.replace('.', ',').rsplit(',', 1)
    except ValueError:
        h_m_s, milliseconds = ts.rsplit(':', 1)

    seconds = reduce(lambda sum, d: sum * 60 + int(d), h_m_s.split(':'), 0)

    return (seconds * 1000) + int(milliseconds)


def get_srt_duration(srt_contents: str, default_end_seconds=0.0) -> float:
    """
    Gets the total duration (based on end timestamp) of an SRT file
    """
    caption_text = srt_contents.split('\n')
    captions_end_seconds = default_end_seconds
    following_line = ''

    for line in reversed(caption_text):
        if '-->' in line:
            # Fix: sometimes the durations will be listed for
            # a blank line (no dialogue)
            if not following_line.strip():
                continue

            end = line.replace(' ', '').rsplit('-->', 1)[-1]
            captions_end_seconds = float(total_seconds(end))
            break

        following_line = line

    return captions_end_seconds



"""
A Python wrapper library for the Wistia API


Docs:
    - https://wistia.com/support/developers/data-api
    - https://wistia.com/support/developers/upload-api
    - https://wistia.com/support/developers/embed-options

"""
__all__ = ['WistiaApi',
           'WistiaDataApi',
           'WistiaEmbedApi',
           'WistiaUploadApi',
           'WistiaHelper']

import logging

from .api_data import WistiaDataApi
from .api_embed import WistiaEmbedApi
from .api_upload import WistiaUploadApi
from .api_helper import WistiaHelper


# A handy alias in case it comes in useful to anyone :-)
WistiaApi = WistiaDataApi


__author__ = 'Ritvik Nag'
__email__ = 'rv.kvetch@gmail.com'
__version__ = '1.0.0'


# Set up logging to ``/dev/null`` like a library is supposed to.
#   https://docs.python.org/3.8/howto/logging.html#configuring-logging-for-a-library
logging.getLogger('wystia').addHandler(logging.NullHandler())



"""
Project-specific constants
"""

__all__ = ['LOG_LEVEL',
           'WISTIA_ACCOUNT',
           'WISTIA_API_TOKEN',
           'RAISE_ON_UNKNOWN_KEY']

import os

from dataclass_wizard.utils.type_conv import as_bool


# Library Log Level
LOG_LEVEL = os.getenv('WISTIA_LOG_LEVEL', 'WARNING').upper()

# This is also the wistia.com sub-domain where videos are hosted
WISTIA_ACCOUNT = os.getenv('WISTIA_ACCOUNT', 'REPLACE-ME')

# Wistia API token
WISTIA_API_TOKEN = os.getenv('WISTIA_API_TOKEN')

# Whether to raise when an unknown JSON key is encountered in the
# de-serialization process; default to true.
RAISE_ON_UNKNOWN_KEY = as_bool(os.getenv('RAISE_ON_UNKNOWN_KEY', True))



from __future__ import annotations

from functools import partial

import requests
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .requests_config import (
    DEFAULT_MAX_RETRIES, DEFAULT_BACKOFF_FACTOR, DEFAULT_STATUS_FORCE_LIST)


class SessionWithRetry(Session):
    """
    Extend the :class:`request.Session` class to provide support for HTTP
    retries (in cases of timeout or other server-side errors)
    """
    def __init__(self, auth=None,
                 num_retries=DEFAULT_MAX_RETRIES,
                 backoff_factor=DEFAULT_BACKOFF_FACTOR,
                 additional_status_force_list: list[int] | None = None):

        super().__init__()
        self.auth = auth

        status_force_list = DEFAULT_STATUS_FORCE_LIST
        # Retry on additional status codes (ex. HTTP 400) if needed
        if additional_status_force_list:
            status_force_list.extend(additional_status_force_list)

        retry_strategy = Retry(
            read=0,
            total=num_retries,
            status_forcelist=status_force_list,
            allowed_methods=["HEAD", "GET", "PUT", "POST", "DELETE",
                             "OPTIONS", "TRACE"],
            backoff_factor=backoff_factor
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)

        self.mount("https://", adapter)
        self.mount("http://", adapter)


def prefix_url_session(prefix: str, session=None) -> Session:
    """
    Returns a :class:`requests.Session` object which makes all HTTP requests
    against a base API endpoint.

    https://stackoverflow.com/a/53140699/10237506
    """
    prefix = prefix.rstrip('/') + '/'

    def new_request(prefix, f, method, url, *args, **kwargs):
        return f(method, prefix + url.lstrip('/'), *args, **kwargs)

    if session is None:
        session = requests.Session()

    session.request = partial(new_request, prefix, session.request)
    return session



"""
Config for retries using the `requests` library
"""


# Default timeout for requests
DEFAULT_TIMEOUT = 60

# Default backoff factor to use for the exponential backoff algorithm;
# Note that the first retry will happen immediately.
#
# For example, if the backoff factor is 1, then after the first retry,
# we will sleep using the following seconds between retries:
#     [2, 4, 8, 16, 32 ...]
DEFAULT_BACKOFF_FACTOR = 1

# Default max retries before giving up.
DEFAULT_MAX_RETRIES = 7

# Maximum backoff time.
BACKOFF_MAX = 120

# Set of HTTP status codes that we should force a retry on
DEFAULT_STATUS_FORCE_LIST = [429, 500, 502, 503, 504]



from __future__ import annotations

import os.path
from typing import Any

from requests import HTTPError
from requests_toolbelt import MultipartEncoder

from .api_base import _BaseWistiaApi
from .config import WistiaConfig
from .errors import UploadFailed
from .log import LOG
from .models import UploadResponse
from .utils.decorators import retry_on_connection_error


class WistiaUploadApi(_BaseWistiaApi):
    """
    Helper class to interact with the Wistia Upload API (docs below)
      https://wistia.com/support/developers/upload-api

    """
    _API_ENDPOINT = WistiaConfig.UPLOAD_URL

    @classmethod
    def upload_file(
        cls,
        file_path: str,
        project_id: str | None = None,
        title: str | None = None,
        description: str | None = None,
        contact_id: int | None = None,
        max_retries=5
    ) -> UploadResponse:
        """
        Uploads a video file (given an absolute file path) to Wistia,
        and returns the hashed ID of the newly hosted video.

        Docs:
          https://wistia.com/support/developers/upload-api#the-request

        :param file_path: The path of the video file to upload.
        :param project_id: The hashed id of the project to upload media into.
          If omitted, a new project will be created and uploaded to.
        :param title: Optional display name for the video. If omitted, the
          filename will be used instead.
        :param description: Optional description for the video.
        :param contact_id: A Wistia contact id, an integer value. If omitted,
          it will default to the contact_id of the account’s owner.
        :param max_retries: Maximum number of retries, in case we run into a
          `BrokenPipeError` while uploading the video; defaults to 5.
        :raises UploadFailed: In case the upload operation fails; also logs
          error details from the Upload API response.

        """
        file_data = {'file': open(file_path, 'rb'),
                     'name': title or os.path.basename(file_path),
                     'access_token': cls._API_TOKEN}
        if project_id:
            file_data['project_id'] = project_id
        if description:
            file_data['description'] = description
        if contact_id:
            file_data['contact_id'] = contact_id

        # Works around a :class:`OverflowError` that is usually raised by the
        # `requests` library when uploading large files (>2GB).
        m = MultipartEncoder(fields=file_data)

        upload_func = retry_on_connection_error(
            cls._upload_url_or_file_to_wistia, max_retries=max_retries)
        data = upload_func(m, {'Content-Type': m.content_type})

        return UploadResponse.from_dict(data)

    @classmethod
    def upload_link(
        cls,
        url: str,
        project_id: str | None = None,
        title: str | None = None,
        description: str | None = None,
        contact_id: int | None = None
    ) -> UploadResponse:
        """
        Uploads a public video link to Wistia, and returns the hashed ID of the
        newly hosted video.

        Docs:
          https://wistia.com/support/developers/upload-api#the-request

        :param url: A public, downloadable link for the video.
        :param project_id: The hashed id of the project to upload media into.
          If omitted, a new project will be created and uploaded to.
        :param title: Optional display name for the video. If omitted, the
          filename will be used instead.
        :param description: Optional description for the video.
        :param contact_id: A Wistia contact id, an integer value. If omitted,
          it will default to the contact_id of the account’s owner.
        :raises UploadFailed: In case the upload operation fails; also logs
          error details from the Upload API response.

        """
        file_data = {'url': url,
                     'name': title or os.path.basename(url),
                     'access_token': cls._API_TOKEN}
        if project_id:
            file_data['project_id'] = project_id
        if description:
            file_data['description'] = description
        if contact_id:
            file_data['contact_id'] = contact_id

        data = cls._upload_url_or_file_to_wistia(file_data)

        return UploadResponse.from_dict(data)

    @classmethod
    def _upload_url_or_file_to_wistia(
            cls, data, headers=None) -> dict[str, Any]:
        """
        Passes a URL or a file to the Wistia Upload API.
        """
        # Make the API request to upload the video
        LOG.debug('Uploading to Wistia...')
        r = cls._get_session().request(
            'POST', '/', data=data, headers=headers)
        try:
            # Confirm the request was a success
            r.raise_for_status()
        except HTTPError:
            # Raise an error with the response details
            raise UploadFailed(r)

        LOG.info('Upload successful, completed in %s', r.elapsed)
        return r.json()



"""
Unit Tests for the `wystia` package.
"""
import logging
from unittest.mock import Mock

import pytest

from wystia import *
from wystia.api_base import _BaseWistiaApi
from wystia.errors import *
from wystia.models import *
from conftest import PY36

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def test_configure_works_as_expected(mock_open, mock_api_token):
    """
    Test that the :attr:`_API_TOKEN` class attribute is shared
    between sibling classes that share the same base class.

    """
    for cls in _BaseWistiaApi, WistiaDataApi, WistiaUploadApi:
        assert cls._API_TOKEN is None

    WistiaDataApi.configure(mock_api_token)

    for cls in _BaseWistiaApi, WistiaDataApi, WistiaUploadApi:
        assert cls._API_TOKEN is not None
        assert cls._API_TOKEN is mock_api_token

    WistiaDataApi.configure(None)

    for cls in _BaseWistiaApi, WistiaDataApi, WistiaUploadApi:
        assert cls._API_TOKEN is None


def test_request_count_is_shared(
    mock_from_dict,
    mock_open,
    mock_video_id,
    mock_file_name
):
    """
    Test that the :attr:`_REQUEST_COUNT` class attribute is shared
    between sibling classes that share the same base class.

    """
    r = WistiaDataApi.get_video(mock_video_id)
    assert isinstance(r, Mock)

    r = WistiaUploadApi.upload_file(mock_file_name)
    assert isinstance(r, Mock)

    for cls in WistiaDataApi, WistiaUploadApi:
        assert cls.request_count() == 2, \
            'Wrong request count for class: ' + cls.__name__

    WistiaDataApi.reset_request_count()

    for cls in WistiaDataApi, WistiaUploadApi:
        assert cls.request_count() == 0, \
            'Request count not reset for class: ' + cls.__name__


def test_wistia_error_logs_with_expected_kwargs(mock_log, mock_video_id):
    _ = NoSuchMedia(mock_video_id)

    mock_log.error.assert_called_with(
        '%s: %s',
        'NoSuchMedia',
        'Video (or media) does not exist, or was deleted from Wistia. media_id={}'.format(
            mock_video_id))


def test_create_captions_when_no_content(mock_video_id, mock_log):
    """Test case for `create_captions` when captions are not provided."""
    with pytest.raises(ContentIsEmpty):
        _ = WistiaDataApi.create_captions(
            mock_video_id)

    if PY36:
        mock_log.error.assert_called_once()


def test_update_captions_when_no_content(mock_video_id, mock_log):
    """Test case for `update_captions` when captions are not provided."""
    with pytest.raises(ContentIsEmpty):
        _ = WistiaDataApi.update_captions(
            mock_video_id, LanguageCode.ENGLISH)

    if PY36:
        mock_log.error.assert_called_once()


def test_is_archived_video():
    assert WistiaHelper.is_archived_video(
        'My Video Title [Archived on August 13, 2015]')


def test_video_data_methods():
    vd = Video.from_dict(
        {
            'type': 'Video',
            'duration': 7.2,
            "hashedId": "abc12345",
            "id": 987654321,
            "name": "My Video",
            "description": "Some Description",
            "created": "2020-10-22T16:07:24Z",
            "updated": "2021-12-14T14:09:59Z",
            "progress": 1.0,
            "thumbnail": {
                "url": "https://embed-ssl.wistia.com/deliveries/abc.jpg",
                "width": 200,
                "height": 120
            }
        }
    )

    assert vd.duration == 7.2
    assert vd.status == MediaStatus.NOT_FOUND
    assert not vd.ad_disabled

    log.debug('Video Data object: %s', vd)
    assert repr(vd).startswith(Video.__name__ + '(')

    log.debug('Video Data dictionary result: %s', vd.to_dict())


def test_video_embed_data_methods(mock_video_id):
    video_name = 'my-video.mp4'
    created_ts = 1234567
    duration = 7.2
    my_src_url = "https://embed-ssl.wistia.com/deliveries/abc.bin"

    ved = VideoEmbedData.from_dict({'type': 'Video',
                                    'hashedId': mock_video_id,
                                    'name': video_name,
                                    'createdAt': created_ts,
                                    'duration': duration,
                                    "assets": [
                                        {
                                            "type": "original",
                                            "slug": "original",
                                            "displayName": "Original file",
                                            "bitrate": 2356,
                                            "public": True,
                                            "status": 2,
                                            "progress": 1.0,
                                            "url": my_src_url,
                                            "createdAt": "2021-12-29T16:34:39",
                                            "size": 132180888,
                                            "metadata": {
                                                "servedByMediaApi": 1
                                            },
                                            "width": 1920,
                                            "height": 1080
                                        }
                                    ],
                                    "projectId": 1234567,
                                    "stats": {
                                        "loadCount": 3128,
                                        "playCount": 2011,
                                        "uniqueLoadCount": 2502,
                                        "uniquePlayCount": 2111,
                                        "averageEngagement": 0.1
                                    },
                                    "distilleryUrl": "https://distillery.wistia.com/x",
                                    "accountKey": "wistia-production_3210",
                                    "mediaKey": "wistia-production_112233",
                                    "mediaType": "Video",
                                    "progress": 1.0,
                                    "status": 2,
                                    "branding": False,
                                    "enableCustomerLogo": True,
                                    "seoDescription": "sample description",
                                    "preloadPreference": None,
                                    "flashPlayerUrl": "https://embed-ssl.wistia.com/flash/embed_player_v2.0.swf",
                                    "showAbout": True,
                                    "firstEmbedForAccount": False,
                                    "firstShareForAccount": False,
                                    "keyframeAlign": True,
                                    "useMediaDataHostLogic": True,
                                    "trackingTransmitInterval": 20,
                                    "integrations": {},
                                    "hlsEnabled": True,
                                    "embedOptions": {
                                        "volumeControl": True,
                                        "fullscreenButton": True,
                                        "controlsVisibleOnLoad": True,
                                        "playerColor": "12345",
                                        "bpbTime": False,
                                        "plugin": {
                                            "captionsV1": {
                                                "on": True
                                            }
                                        },
                                        "vulcan": True,
                                        "playsinline": True,
                                    }
                                    })

    assert ved.name == video_name
    assert ved.hashed_id == mock_video_id
    assert ved.duration == duration
    assert ved.created_at.timestamp() == created_ts
    assert '/abc/' in ved.source_url

    log.debug('Video Embed Data object: %r', ved)
    assert repr(ved).startswith(VideoEmbedData.__name__ + '(')


def test_upload_response_methods(mock_video_id):
    video_name = 'my-video.mp4'
    description = 'my desc'
    created_string = '2021-01-02T14:30:59'
    status = 'queued'
    duration = 7.2

    ur = UploadResponse.from_dict(
        {'type': 'Video',
         'id': 2208087,
         'hashed_id': mock_video_id,
         'name': video_name,
         'description': description,
         'created': created_string,
         'updated': created_string,
         'progress': '0',
         'status': status,
         'duration': duration,
         'thumbnail':
             {
                 'url': 'http://embed.wistia.com/deliveries/abc.jpg',
                 'width': 100,
                 'height': 60
             },
         'accountId': 12345
         },
    )

    assert ur.name == video_name
    assert ur.description == description
    assert ur.hashed_id == mock_video_id
    assert ur.duration == duration
    assert ur.status == MediaStatus(status)
    assert ur.created.isoformat() == created_string
    assert not ur.progress  # eq: 0

    log.debug('Video Upload Response object: %r', ur)
    assert repr(ur).startswith(UploadResponse.__name__ + '(')



"""
Configs and settings for Wistia API
"""
from .constants import WISTIA_ACCOUNT


class WistiaConfig:

    API_URL = 'https://api.wistia.com/v1/'
    UPLOAD_URL = 'https://upload.wistia.com/'
    EMBED_URL = 'https://fast.wistia.com/embed/'

    PROJECTS_URL = 'projects.json'
    PROJECTS_SHOW_URL = 'projects/{project_id}.json'
    PROJECTS_COPY_URL = 'projects/{project_id}/copy.json'
    MEDIAS_URL = 'medias.json'
    MEDIAS_SHOW_URL = 'medias/{media_id}.json'
    MEDIAS_COPY_URL = 'medias/{media_id}/copy.json'
    MEDIAS_STATS_URL = 'medias/{media_id}/stats.json'
    CUSTOMIZATION_URL = 'medias/{media_id}/customizations.json'
    CAPTIONS_ORDER_URL = 'medias/{media_id}/captions/purchase.json'
    ALL_CAPTIONS_URL = 'medias/{media_id}/captions.json'
    LANG_CAPTIONS_URL = 'medias/{media_id}/captions/{lang_code}.json'

    MEDIAS_EMBED_URL = 'medias/{media_id}.json'

    @classmethod
    def wistia_url(cls, video_id, account_name=WISTIA_ACCOUNT):
        """Construct the wistia media Url given a video id"""
        return 'https://{account_name}.wistia.com/medias/{video_id}'.format(
            account_name=account_name, video_id=video_id)



"""
Project-specific exception classes
"""
__all__ = ['WistiaError',
           'ContentIsEmpty',
           'NoSuchProject',
           'NoSuchMedia',
           'VideoHasCaptions',
           'UploadFailed']

from requests import Response

from .log import LOG
from .utils.response import format_error


class WistiaError(Exception):
    """
    Base exception class for errors raised by this library.

    """
    ERR_STATUS = 400

    def __init__(self, message, **log_kwargs):
        self.message = message
        self.code = self.__class__.__name__

        super(WistiaError, self).__init__(self.message)

        if log_kwargs:
            field_vals = [f'{k}={v}' for k, v in log_kwargs.items()]
            message = f'{message.rstrip(".")}. {", ".join(field_vals)}'

        LOG.error('%s: %s', self.code, message)

    def response(self):
        """Formats an error object and returns an AWS Lambda Proxy response."""
        return format_error(self.message, self.code, self.ERR_STATUS)


class ContentIsEmpty(WistiaError):

    def __init__(self):
        msg = 'The specified file content in the request is empty.'
        super(ContentIsEmpty, self).__init__(msg)


class NoSuchProject(WistiaError):

    def __init__(self, project_id):
        msg = 'Project does not exist, or was deleted from Wistia.'
        super(NoSuchProject, self).__init__(msg, project_id=project_id)


class NoSuchMedia(WistiaError):

    def __init__(self, media_id):
        msg = 'Video (or media) does not exist, or was deleted from Wistia.'
        super(NoSuchMedia, self).__init__(msg, media_id=media_id)


class VideoHasCaptions(WistiaError):

    def __init__(self, video_id):
        msg = 'English captions already exist for the Wistia video.'
        super(VideoHasCaptions, self).__init__(msg, video_id=video_id)


class UploadFailed(WistiaError):

    def __init__(self, r: Response):
        msg = 'Failure calling the Wistia Upload API.'
        super(UploadFailed, self).__init__(
            msg, status=r.status_code, text=r.text, reason=r.reason)



"""The setup script."""
import pathlib

from setuptools import setup


here = pathlib.Path(__file__).parent

package_name = 'wystia'

packages = [
    package_name,
    f'{package_name}.utils',
    f'{package_name}.utils.parse'
]

requires = [
    'requests',
    'requests-toolbelt',
    'urllib3',  # should already be installed via requests
    'dataclass-wizard>=0.18.0,<1.0',
    'cached-property~=1.5.2; python_version == "3.7"'
]

test_requirements = [
    'pytest>=6',
    'pytest-mock~=3.6.1'
]

readme = (here / 'README.rst').read_text()
history = (here / 'HISTORY.rst').read_text()

setup(
    name='wystia',
    version='1.0.0',
    description='A Python wrapper library for the Wistia API',
    long_description=readme + '\n\n' + history,
    author='Ritvik Nag',
    author_email='rv.kvetch@gmail.com',
    url='https://github.com/rnag/wystia',
    packages=packages,
    include_package_data=True,
    python_requires='>=3.7',
    install_requires=requires,
    license='MIT',
    keywords=['wistia', 'wistia api', 'wystia',
              'wistia data api', 'wistia upload api'],
    classifiers=[
        # Ref: https://pypi.org/classifiers/
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python'
    ],
    test_suite='tests',
    tests_require=test_requirements,
    zip_safe=False
)



import pytest

from wystia.utils.parse.srt import *


@pytest.mark.parametrize(
    'input,expected',
    [
        ('1:20:32.172', '4832.172'),
        ('01:23:45,5', '5025.005'),
        ('0:07:00:1', '420.001'),
        ('3:01:10:000', '10870.000'),
        ('12:57:42:500', '46662.500'),
    ]
)
def test_total_seconds(input, expected):
    actual = total_seconds(input)
    assert actual == expected, 'Total seconds not as expected'


@pytest.mark.parametrize(
    'input,expected',
    [
        ('1:20:32.172', 4832172),
        ('01:23:45,5', 5025005),
        ('0:07:00:1', 420001),
        ('3:01:10:000', 10870000),
        ('12:57:42:500', 46662500),
    ]
)
def test_total_ms(input, expected):
    actual = total_ms(input)
    assert actual == expected, 'Total seconds not as expected'


@pytest.mark.parametrize(
    'srt_contents,default_end_seconds,expected',
    [
        (
            """
            1
            00:00:01,250 --> 00:07:07,250
            Testing.

            1
            00:07:07,250 --> 00:07:10,160
            Hello world!

            4
            00:07:13,880 --> 00:07:15,950
            Another value.
            """, 0.5, 435.95
        ),
        (
            """
            1
            Invalid srt contents here
            """, 0.5, 0.5
        ),
        (
            """
            1
            00:00:01,250 --> 00:03:00,000
            Testing.

            2
            00:03:00,020 --> 00:03:00,040
            """, 0.5, 180.0
        ),
    ],
    # Attempt to shorten the test case names that are displayed
    # Ref: https://docs.pytest.org/en/stable/example/parametrize.html#different-options-for-test-ids    # noqa: E501
    ids=['Valid SRT', 'Invalid SRT', 'Skip Line with Empty Dialogue']
)
def test_get_srt_duration(srt_contents, default_end_seconds, expected):
    actual = get_srt_duration(srt_contents, default_end_seconds)
    assert actual == expected



"""
Metaclass utilities
"""
import pprint


def display_with_pformat(name: str, bases: tuple, cls_dict: dict):
    """
    Metaclass which overrides the :meth:`__str__` of a class to use the
    `pprint` module to pretty-format the class's default :meth:`__repr__`.
    """
    def __str__(self):
        return pprint.pformat(self)

    cls_dict['__str__'] = __str__

    return type(name, bases, cls_dict)
